@@ -12,6 +12,7 @@
 #include <stdlib.h>     // for free
 #include <sys/types.h>  // for ssize_t
 
+#include <limits>
 #include <string>  // for string, u16string
 #include <type_traits>  // for enable_if_t, add_pointer_t, add_const_t
 #include <utility>      // IWYU pragma: keep
@@ -21,6 +22,7 @@
 #include <glib-object.h>
 #include <glib.h>
 
+#include <js/BigInt.h>
 #include <js/GCAPI.h>
 #include <js/GCPolicyAPI.h>  // for IgnoreGCPolicy
 #include <js/Id.h>
@@ -29,6 +31,11 @@
 #include <jspubtd.h>     // for JSProtoKey
 
 #include "cjs/macros.h"
+#include "util/log.h"
+
+#if GJS_VERBOSE_ENABLE_MARSHAL
+#    include "gi/arg-types-inl.h"  // for static_type_name
+#endif
 
 class JSErrorReport;
 namespace JS {
@@ -109,12 +116,12 @@ struct GjsAutoPointer {
         return *this;
     }
 
-    GjsAutoPointer& operator=(GjsAutoPointer&& other) {
+    constexpr GjsAutoPointer& operator=(GjsAutoPointer&& other) {
         this->swap(other);
         return *this;
     }
 
-    GjsAutoPointer& operator=(GjsAutoPointer const& other) {
+    constexpr GjsAutoPointer& operator=(GjsAutoPointer const& other) {
         GjsAutoPointer dup(other.get(), GjsAutoTakeOwnership());
         this->swap(dup);
         return *this;
@@ -139,6 +146,7 @@ struct GjsAutoPointer {
 
     constexpr Ptr get() const { return m_ptr; }
     constexpr Ptr* out() { return &m_ptr; }
+    constexpr ConstPtr* out() const { return &m_ptr; }
 
     constexpr Ptr release() {
         auto* ptr = m_ptr;
@@ -151,12 +159,8 @@ struct GjsAutoPointer {
         m_ptr = ptr;
 
         if constexpr (has_free_function()) {
-            if (old_ptr) {
-                if constexpr (std::is_array_v<T>)
-                    free_func(reinterpret_cast<T*>(old_ptr));
-                else
-                    free_func(old_ptr);
-            }
+            if (old_ptr)
+                free_func(reinterpret_cast<F*>(old_ptr));
         }
     }
 
@@ -172,7 +176,9 @@ struct GjsAutoPointer {
     [[nodiscard]] constexpr std::enable_if_t<!std::is_array_v<U>, Ptr> copy()
         const {
         static_assert(has_ref_function(), "No ref function provided");
-        return m_ptr ? reinterpret_cast<Ptr>(ref_func(m_ptr)) : nullptr;
+        return m_ptr ? reinterpret_cast<Ptr>(
+                           ref_func(reinterpret_cast<F*>(m_ptr)))
+                     : nullptr;
     }
 
     template <typename C>
@@ -442,12 +448,9 @@ bool        gjs_log_exception                (JSContext       *context);
 
 bool gjs_log_exception_uncaught(JSContext* cx);
 
-bool gjs_log_exception_full(JSContext* cx, JS::HandleValue exc,
+void gjs_log_exception_full(JSContext* cx, JS::HandleValue exc,
                             JS::HandleString message, GLogLevelFlags level);
 
-[[nodiscard]] std::string gjs_value_debug_string(JSContext* cx,
-                                                 JS::HandleValue value);
-
 void gjs_warning_reporter(JSContext*, JSErrorReport* report);
 
 GJS_JSAPI_RETURN_CONVENTION
@@ -562,6 +565,7 @@ bool gjs_object_require_converted_property(JSContext       *context,
                                            JS::HandleId     property_name,
                                            uint32_t        *value);
 
+[[nodiscard]] std::string gjs_debug_bigint(JS::BigInt* bi);
 [[nodiscard]] std::string gjs_debug_string(JSString* str);
 [[nodiscard]] std::string gjs_debug_symbol(JS::Symbol* const sym);
 [[nodiscard]] std::string gjs_debug_object(JSObject* obj);
@@ -601,6 +605,36 @@ static constexpr size_t N_REASONS = 0 FOREACH_GC_REASON(COUNT_GC_REASON);
 #undef COUNT_GC_REASON
 };
 
+template <typename T>
+[[nodiscard]] bool bigint_is_out_of_range(JS::BigInt* bi, T* clamped) {
+    static_assert(sizeof(T) == 8, "64-bit types only");
+    g_assert(bi && "bigint cannot be null");
+    g_assert(clamped && "forgot out parameter");
+
+    gjs_debug_marshal(GJS_DEBUG_GFUNCTION,
+                      "Checking if BigInt %s is out of range for type %s",
+                      gjs_debug_bigint(bi).c_str(), Gjs::static_type_name<T>());
+
+    if (JS::BigIntFits(bi, clamped)) {
+        gjs_debug_marshal(
+            GJS_DEBUG_GFUNCTION, "BigInt %s is in the range of type %s",
+            std::to_string(*clamped).c_str(), Gjs::static_type_name<T>());
+        return false;
+    }
+
+    if (JS::BigIntIsNegative(bi)) {
+        *clamped = std::numeric_limits<T>::min();
+    } else {
+        *clamped = std::numeric_limits<T>::max();
+    }
+
+    gjs_debug_marshal(GJS_DEBUG_GFUNCTION,
+                      "BigInt %s is not in the range of type %s, clamped to %s",
+                      gjs_debug_bigint(bi).c_str(), Gjs::static_type_name<T>(),
+                      std::to_string(*clamped).c_str());
+    return true;
+}
+
 }  // namespace Gjs
 
 [[nodiscard]] const char* gjs_explain_gc_reason(JS::GCReason reason);
