@@ -6,56 +6,69 @@
 
 #include <stddef.h>     // for size_t
 #include <string.h>
-#include <sys/types.h>  // for ssize_t
 
 #include <string>  // for u16string
 
 #include <gio/gio.h>
 #include <glib.h>
 
+#include <js/CallAndConstruct.h>
 #include <js/CallArgs.h>
 #include <js/CharacterEncoding.h>  // for ConstUTF8CharsZ
 #include <js/Class.h>
 #include <js/CompilationAndEvaluation.h>
 #include <js/CompileOptions.h>
 #include <js/Conversions.h>
+#include <js/ErrorReport.h>  // for JS_ReportOutOfMemory
+#include <js/Exception.h>
 #include <js/GCVector.h>  // for RootedVector
+#include <js/GlobalObject.h>  // for CurrentGlobalOrNull
 #include <js/Id.h>
 #include <js/Modules.h>
+#include <js/Object.h>
 #include <js/Promise.h>
+#include <js/PropertyAndElement.h>
 #include <js/PropertyDescriptor.h>
+#include <js/Realm.h>
 #include <js/RootingAPI.h>
+#include <js/ScriptPrivate.h>
 #include <js/SourceText.h>
+#include <js/String.h>
 #include <js/TypeDecls.h>
 #include <js/Utility.h>  // for UniqueChars
 #include <js/Value.h>
 #include <js/ValueArray.h>
-#include <jsapi.h>  // for JS_DefinePropertyById, ...
-#include <jsfriendapi.h>  // for SetFunctionNativeReserved
+#include <jsapi.h>        // for JS_GetFunctionObject, JS_Ne...
+#include <jsfriendapi.h>  // for NewFunctionWithReserved
+#include <mozilla/Maybe.h>
 
 #include "cjs/atoms.h"
 #include "cjs/context-private.h"
 #include "cjs/global.h"
 #include "cjs/jsapi-util-args.h"
 #include "cjs/jsapi-util.h"
+#include "cjs/macros.h"
 #include "cjs/mem-private.h"
 #include "cjs/module.h"
 #include "cjs/native.h"
 #include "util/log.h"
 #include "util/misc.h"
 
-class GjsScriptModule {
-    char *m_name;
+namespace mozilla {
+union Utf8Unit;
+}
 
-    GjsScriptModule(const char* name) {
-        m_name = g_strdup(name);
+class GjsScriptModule {
+    GjsAutoChar m_name;
+
+    // Reserved slots
+    static const size_t POINTER = 0;
+
+    GjsScriptModule(const char* name) : m_name(g_strdup(name)) {
         GJS_INC_COUNTER(module);
     }
 
-    ~GjsScriptModule() {
-        g_free(m_name);
-        GJS_DEC_COUNTER(module);
-    }
+    ~GjsScriptModule() { GJS_DEC_COUNTER(module); }
 
     GjsScriptModule(GjsScriptModule&) = delete;
     GjsScriptModule& operator=(GjsScriptModule&) = delete;
@@ -63,13 +76,15 @@ class GjsScriptModule {
     /* Private data accessors */
 
     [[nodiscard]] static inline GjsScriptModule* priv(JSObject* module) {
-        return static_cast<GjsScriptModule*>(JS_GetPrivate(module));
+        return JS::GetMaybePtrFromReservedSlot<GjsScriptModule>(
+            module, GjsScriptModule::POINTER);
     }
 
     /* Creates a JS module object. Use instead of the class's constructor */
     [[nodiscard]] static JSObject* create(JSContext* cx, const char* name) {
         JSObject* module = JS_NewObject(cx, &GjsScriptModule::klass);
-        JS_SetPrivate(module, new GjsScriptModule(name));
+        JS::SetReservedSlot(module, GjsScriptModule::POINTER,
+                            JS::PrivateValue(new GjsScriptModule(name)));
         return module;
     }
 
@@ -84,7 +99,7 @@ class GjsScriptModule {
         if (!JS_DefinePropertyById(cx, importer, name, module,
                                    GJS_MODULE_PROP_FLAGS & ~JSPROP_PERMANENT)) {
             gjs_debug(GJS_DEBUG_IMPORTER, "Failed to define '%s' in importer",
-                      m_name);
+                      m_name.get());
             return false;
         }
 
@@ -94,22 +109,10 @@ class GjsScriptModule {
     /* Carries out the actual execution of the module code */
     GJS_JSAPI_RETURN_CONVENTION
     bool evaluate_import(JSContext* cx, JS::HandleObject module,
-                         const char* script, ssize_t script_len,
+                         const char* source, size_t source_len,
                          const char* filename, const char* uri) {
-        long items_written;  // NOLINT(runtime/int) - required by GLib API
-        GError* error;
-        GjsAutoChar16 utf16_string =
-            g_utf8_to_utf16(script, script_len,
-                            /* items_read = */ nullptr, &items_written, &error);
-        if (!utf16_string)
-            return gjs_throw_gerror_message(cx, error);
-
-        // COMPAT: This could use JS::SourceText<mozilla::Utf8Unit> directly,
-        // but that messes up code coverage. See bug
-        // https://bugzilla.mozilla.org/show_bug.cgi?id=1404784
-        JS::SourceText<char16_t> buf;
-        if (!buf.init(cx, reinterpret_cast<char16_t*>(utf16_string.get()),
-                      items_written, JS::SourceOwnership::Borrowed))
+        JS::SourceText<mozilla::Utf8Unit> buf;
+        if (!buf.init(cx, source, source_len, JS::SourceOwnership::Borrowed))
             return false;
 
         JS::RootedObjectVector scope_chain(cx);
@@ -119,19 +122,26 @@ class GjsScriptModule {
         }
 
         JS::CompileOptions options(cx);
-        options.setFileAndLine(filename, 1);
+        options.setFileAndLine(filename, 1).setNonSyntacticScope(true);
 
         JS::RootedObject priv(cx, build_private(cx, uri));
-        options.setPrivateValue(JS::ObjectValue(*priv));
+        if (!priv)
+            return false;
 
+        JS::RootedScript script(cx, JS::Compile(cx, options, buf));
+        if (!script)
+            return false;
+
+        JS::SetScriptPrivate(script, JS::ObjectValue(*priv));
         JS::RootedValue ignored_retval(cx);
-        if (!JS::Evaluate(cx, scope_chain, options, buf, &ignored_retval))
+        if (!JS_ExecuteScript(cx, scope_chain, script, &ignored_retval))
             return false;
 
         GjsContextPrivate* gjs = GjsContextPrivate::from_cx(cx);
         gjs->schedule_gc_if_needed();
 
-        gjs_debug(GJS_DEBUG_IMPORTER, "Importing module %s succeeded", m_name);
+        gjs_debug(GJS_DEBUG_IMPORTER, "Importing module %s succeeded",
+                  m_name.get());
 
         return true;
     }
@@ -144,14 +154,12 @@ class GjsScriptModule {
                 GFile           *file)
     {
         GError *error = nullptr;
-        char *unowned_script;
+        GjsAutoChar script;
         size_t script_len = 0;
 
-        if (!(g_file_load_contents(file, nullptr, &unowned_script, &script_len,
+        if (!(g_file_load_contents(file, nullptr, script.out(), &script_len,
                                    nullptr, &error)))
             return gjs_throw_gerror_message(cx, error);
-
-        GjsAutoChar script = unowned_script;  /* steals ownership */
         g_assert(script);
 
         GjsAutoChar full_path = g_file_get_parse_name(file);
@@ -174,27 +182,29 @@ class GjsScriptModule {
             return true;  /* nothing imported yet */
         }
 
-        if (!JS_HasPropertyById(cx, lexical, id, resolved))
+        JS::Rooted<mozilla::Maybe<JS::PropertyDescriptor>> maybe_desc(cx);
+        JS::RootedObject holder(cx);
+        if (!JS_GetPropertyDescriptorById(cx, lexical, id, &maybe_desc,
+                                          &holder))
             return false;
-        if (!*resolved)
+        if (maybe_desc.isNothing())
             return true;
 
         /* The property is present in the lexical environment. This should not
          * be supported according to ES6. For compatibility with earlier GJS,
          * we treat it as if it were a real property, but warn about it. */
 
-        g_warning("Some code accessed the property '%s' on the module '%s'. "
-                  "That property was defined with 'let' or 'const' inside the "
-                  "module. This was previously supported, but is not correct "
-                  "according to the ES6 standard. Any symbols to be exported "
-                  "from a module must be defined with 'var'. The property "
-                  "access will work as previously for the time being, but "
-                  "please fix your code anyway.",
-                  gjs_debug_id(id).c_str(), m_name);
+        g_warning(
+            "Some code accessed the property '%s' on the module '%s'. That "
+            "property was defined with 'let' or 'const' inside the module. "
+            "This was previously supported, but is not correct according to "
+            "the ES6 standard. Any symbols to be exported from a module must "
+            "be defined with 'var'. The property access will work as "
+            "previously for the time being, but please fix your code anyway.",
+            gjs_debug_id(id).c_str(), m_name.get());
 
-        JS::Rooted<JS::PropertyDescriptor> desc(cx);
-        return JS_GetPropertyDescriptorById(cx, lexical, id, &desc) &&
-            JS_DefinePropertyById(cx, module, id, desc);
+        JS::Rooted<JS::PropertyDescriptor> desc(cx, maybe_desc.value());
+        return JS_DefinePropertyById(cx, module, id, desc);
     }
 
     GJS_JSAPI_RETURN_CONVENTION
@@ -207,7 +217,9 @@ class GjsScriptModule {
         return priv(module)->resolve_impl(cx, module, id, resolved);
     }
 
-    static void finalize(JSFreeOp*, JSObject* module) { delete priv(module); }
+    static void finalize(JS::GCContext*, JSObject* module) {
+        delete priv(module);
+    }
 
     static constexpr JSClassOps class_ops = {
         nullptr,  // addProperty
@@ -221,13 +233,14 @@ class GjsScriptModule {
 
     static constexpr JSClass klass = {
         "GjsScriptModule",
-        JSCLASS_HAS_PRIVATE | JSCLASS_BACKGROUND_FINALIZE,
+        JSCLASS_HAS_RESERVED_SLOTS(1) | JSCLASS_BACKGROUND_FINALIZE,
         &GjsScriptModule::class_ops,
     };
 
  public:
     /*
-     * Creates a JS object to pass to JS::CompileOptions as a script's private.
+     * Creates a JS object to pass to JS::SetScriptPrivate as a script's
+     * private.
      */
     GJS_JSAPI_RETURN_CONVENTION
     static JSObject* build_private(JSContext* cx, const char* script_uri) {
@@ -432,7 +445,7 @@ static bool import_native_module_sync(JSContext* cx, unsigned argc,
     }
 
     JS::RootedObject native_obj(cx);
-    if (!gjs_load_native_module(cx, id.get(), &native_obj)) {
+    if (!Gjs::NativeModuleRegistry::get().load(cx, id.get(), &native_obj)) {
         gjs_throw(cx, "Failed to load native module: %s", id.get());
         return false;
     }
@@ -498,7 +511,7 @@ bool gjs_populate_module_meta(JSContext* cx, JS::HandleValue private_ref,
  * @returns whether an error occurred while resolving the specifier.
  */
 JSObject* gjs_module_resolve(JSContext* cx, JS::HandleValue importingModulePriv,
-                             JS::HandleString specifier) {
+                             JS::HandleObject module_request) {
     g_assert((gjs_global_is_type(cx, GjsGlobalType::DEFAULT) ||
               gjs_global_is_type(cx, GjsGlobalType::INTERNAL)) &&
              "gjs_module_resolve can only be called from module-enabled "
@@ -506,6 +519,8 @@ JSObject* gjs_module_resolve(JSContext* cx, JS::HandleValue importingModulePriv,
     g_assert(importingModulePriv.isObject() &&
              "the importing module can't be null, don't add import to the "
              "bootstrap script");
+    JS::RootedString specifier(
+        cx, JS::GetModuleRequestSpecifier(cx, module_request));
 
     JS::RootedObject global(cx, JS::CurrentGlobalOrNull(cx));
     JS::RootedValue v_loader(
@@ -534,28 +549,34 @@ JSObject* gjs_module_resolve(JSContext* cx, JS::HandleValue importingModulePriv,
 // Can fail in JS::FinishDynamicModuleImport(), but will assert if anything
 // fails in fetching the stashed values, since that would be a serious GJS bug.
 GJS_JSAPI_RETURN_CONVENTION
-static bool finish_import(JSContext* cx, const JS::CallArgs& args) {
+static bool finish_import(JSContext* cx, JS::HandleObject evaluation_promise,
+                          const JS::CallArgs& args) {
+    GjsContextPrivate* priv = GjsContextPrivate::from_cx(cx);
+    priv->main_loop_release();
+
     JS::Value callback_priv = js::GetFunctionNativeReserved(&args.callee(), 0);
     g_assert(callback_priv.isObject() && "Wrong private value");
     JS::RootedObject callback_data(cx, &callback_priv.toObject());
 
     JS::RootedValue importing_module_priv(cx);
-    JS::RootedValue v_specifier(cx);
+    JS::RootedValue v_module_request(cx);
     JS::RootedValue v_internal_promise(cx);
     bool ok GJS_USED_ASSERT =
         JS_GetProperty(cx, callback_data, "priv", &importing_module_priv) &&
         JS_GetProperty(cx, callback_data, "promise", &v_internal_promise) &&
-        JS_GetProperty(cx, callback_data, "specifier", &v_specifier);
+        JS_GetProperty(cx, callback_data, "module_request", &v_module_request);
     g_assert(ok && "Wrong properties on private value");
 
-    g_assert(v_specifier.isString() && "Wrong type for specifier");
+    g_assert(v_module_request.isObject() && "Wrong type for module request");
     g_assert(v_internal_promise.isObject() && "Wrong type for promise");
 
-    JS::RootedString specifier(cx, v_specifier.toString());
+    JS::RootedObject module_request(cx, &v_module_request.toObject());
     JS::RootedObject internal_promise(cx, &v_internal_promise.toObject());
 
     args.rval().setUndefined();
-    return JS::FinishDynamicModuleImport(cx, importing_module_priv, specifier,
+
+    return JS::FinishDynamicModuleImport(cx, evaluation_promise,
+                                         importing_module_priv, module_request,
                                          internal_promise);
 }
 
@@ -566,7 +587,7 @@ static bool finish_import(JSContext* cx, const JS::CallArgs& args) {
 GJS_JSAPI_RETURN_CONVENTION
 static bool fail_import(JSContext* cx, const JS::CallArgs& args) {
     if (JS_IsExceptionPending(cx))
-        return finish_import(cx, args);
+        return finish_import(cx, nullptr, args);
     return false;
 }
 
@@ -581,7 +602,7 @@ static bool import_rejected(JSContext* cx, unsigned argc, JS::Value* vp) {
     JS_SetPendingException(cx, args.get(0),
                            JS::ExceptionStackBehavior::DoNotCapture);
 
-    return finish_import(cx, args);
+    return finish_import(cx, nullptr, args);
 }
 
 GJS_JSAPI_RETURN_CONVENTION
@@ -596,15 +617,21 @@ static bool import_resolved(JSContext* cx, unsigned argc, JS::Value* vp) {
     g_assert(args[0].isObject());
     JS::RootedObject module(cx, &args[0].toObject());
 
-    if (!JS::ModuleInstantiate(cx, module) || !JS::ModuleEvaluate(cx, module))
+    JS::RootedValue evaluation_promise(cx);
+    if (!JS::ModuleInstantiate(cx, module) ||
+        !JS::ModuleEvaluate(cx, module, &evaluation_promise))
         return fail_import(cx, args);
 
-    return finish_import(cx, args);
+    g_assert(evaluation_promise.isObject() &&
+             "got weird value from JS::ModuleEvaluate");
+    JS::RootedObject evaluation_promise_object(cx,
+                                               &evaluation_promise.toObject());
+    return finish_import(cx, evaluation_promise_object, args);
 }
 
 bool gjs_dynamic_module_resolve(JSContext* cx,
                                 JS::HandleValue importing_module_priv,
-                                JS::HandleString specifier,
+                                JS::HandleObject module_request,
                                 JS::HandleObject internal_promise) {
     g_assert(gjs_global_is_type(cx, GjsGlobalType::DEFAULT) &&
              "gjs_dynamic_module_resolve can only be called from the default "
@@ -617,10 +644,12 @@ bool gjs_dynamic_module_resolve(JSContext* cx,
         cx, gjs_get_global_slot(global, GjsGlobalSlot::MODULE_LOADER));
     g_assert(v_loader.isObject());
     JS::RootedObject loader(cx, &v_loader.toObject());
+    JS::RootedString specifier(
+        cx, JS::GetModuleRequestSpecifier(cx, module_request));
 
     JS::RootedObject callback_data(cx, JS_NewPlainObject(cx));
     if (!callback_data ||
-        !JS_DefineProperty(cx, callback_data, "specifier", specifier,
+        !JS_DefineProperty(cx, callback_data, "module_request", module_request,
                            JSPROP_PERMANENT) ||
         !JS_DefineProperty(cx, callback_data, "promise", internal_promise,
                            JSPROP_PERMANENT) ||
@@ -647,8 +676,12 @@ bool gjs_dynamic_module_resolve(JSContext* cx,
 
     JS::RootedValue result(cx);
     if (!JS::Call(cx, loader, "moduleResolveAsyncHook", args, &result))
-        return JS::FinishDynamicModuleImport(cx, importing_module_priv,
-                                             specifier, internal_promise);
+        return JS::FinishDynamicModuleImport(cx, nullptr, importing_module_priv,
+                                             module_request, internal_promise);
+
+    // Release in finish_import
+    GjsContextPrivate* priv = GjsContextPrivate::from_cx(cx);
+    priv->main_loop_hold();
 
     JS::RootedObject resolved(
         cx, JS_GetFunctionObject(js::NewFunctionWithReserved(
