@@ -7,18 +7,22 @@
 
 #include <config.h>
 
+#include <stdint.h>
+
+#include <memory>  // for unique_ptr
 #include <unordered_set>
 #include <vector>
 
 #include <ffi.h>
 #include <girepository.h>
+#include <girffi.h>  // for g_callable_info_get_closure_native_address
 #include <glib-object.h>
 #include <glib.h>
 
 #include <js/GCVector.h>
 #include <js/RootingAPI.h>
 #include <js/TypeDecls.h>
-#include <js/Value.h>  // IWYU pragma: keep
+#include <js/Value.h>
 
 #include "gi/closure.h"
 #include "cjs/jsapi-util.h"
@@ -41,34 +45,50 @@ using GjsAutoGClosure =
 
 struct GjsCallbackTrampoline : public Gjs::Closure {
     GJS_JSAPI_RETURN_CONVENTION static GjsCallbackTrampoline* create(
-        JSContext* cx, JS::HandleFunction function,
-        GICallableInfo* callable_info, GIScopeType scope, bool has_scope_object,
-        bool is_vfunc);
+        JSContext* cx, JS::HandleObject callable, GICallableInfo* callable_info,
+        GIScopeType scope, bool has_scope_object, bool is_vfunc);
 
     ~GjsCallbackTrampoline();
 
-    constexpr ffi_closure* closure() const { return m_closure; }
+    void* closure() const {
+#if GI_CHECK_VERSION(1, 71, 0)
+        return g_callable_info_get_closure_native_address(m_info, m_closure);
+#else
+        return m_closure;
+#endif
+    }
+
+    ffi_closure* get_ffi_closure() const {
+        return m_closure;
+    }
+
+    void mark_forever();
+
+    static void prepare_shutdown();
 
  private:
+    ffi_closure* create_closure();
     GJS_JSAPI_RETURN_CONVENTION bool initialize();
-    GjsCallbackTrampoline(JSContext* cx, JS::HandleFunction function,
+    GjsCallbackTrampoline(JSContext* cx, JS::HandleObject callable,
                           GICallableInfo* callable_info, GIScopeType scope,
                           bool has_scope_object, bool is_vfunc);
 
     void callback_closure(GIArgument** args, void* result);
     GJS_JSAPI_RETURN_CONVENTION
     bool callback_closure_inner(JSContext* cx, JS::HandleObject this_object,
-                                JS::MutableHandleValue rval, GIArgument** args,
-                                GITypeInfo* ret_type, int n_args,
-                                int c_args_offset, void* result);
+                                GObject* gobject, JS::MutableHandleValue rval,
+                                GIArgument** args, GITypeInfo* ret_type,
+                                int n_args, int c_args_offset, void* result);
     void warn_about_illegal_js_callback(const char* when, const char* reason);
 
+    static std::vector<GjsAutoGClosure> s_forever_closure_list;
+
     GjsAutoCallableInfo m_info;
     ffi_closure* m_closure = nullptr;
-    std::vector<GjsParamType> m_param_types;
+    std::unique_ptr<GjsParamType[]> m_param_types;
     ffi_cif m_cif;
 
-    GIScopeType m_scope : 2;
+    GIScopeType m_scope : 3;
     bool m_is_vfunc : 1;
 };
 
@@ -81,10 +101,10 @@ class GjsFunctionCallState {
  public:
     std::unordered_set<GIArgument*> ignore_release;
     JS::RootedObject instance_object;
-    JS::RootedValueVector return_values;
+    JS::RootedVector<JS::Value> return_values;
     GjsAutoError local_error;
     GICallableInfo* info;
-    int gi_argc = 0;
+    uint8_t gi_argc = 0;
     unsigned processed_c_args = 0;
     bool failed : 1;
     bool can_throw_gerror : 1;
@@ -115,6 +135,13 @@ class GjsFunctionCallState {
 
     constexpr int first_arg_offset() const { return is_method ? 2 : 1; }
 
+    // The list always contains the return value, and the arguments
+    constexpr GIArgument* instance() {
+        return is_method ? &m_in_cvalues[1] : nullptr;
+    }
+
+    constexpr GIArgument* return_value() { return &m_out_cvalues[0]; }
+
     constexpr GIArgument& in_cvalue(int index) const {
         return m_in_cvalues[index + first_arg_offset()];
     }
@@ -133,6 +160,10 @@ class GjsFunctionCallState {
 
     constexpr bool call_completed() { return !failed && !did_throw_gerror(); }
 
+    constexpr uint8_t last_processed_index() {
+        return first_arg_offset() + processed_c_args;
+    }
+
     [[nodiscard]] GjsAutoChar display_name() {
         GIBaseInfo* container = g_base_info_get_container(info);  // !owned
         if (container) {
