@@ -15,17 +15,23 @@
 #include <glib-object.h>
 #include <glib.h>
 
+#include <js/CallAndConstruct.h>  // for JS_CallFunctionValue
 #include <js/Class.h>
 #include <js/ComparisonOperators.h>
-#include <js/Id.h>                  // for JSID_IS_STRING, JSID_VOID
+#include <js/Exception.h>
+#include <js/GlobalObject.h>        // for CurrentGlobalOrNull
+#include <js/Id.h>                  // for PropertyKey
+#include <js/Object.h>              // for GetClass
+#include <js/PropertyAndElement.h>
 #include <js/PropertyDescriptor.h>  // for JSPROP_PERMANENT, JSPROP_RESOLVING
 #include <js/RootingAPI.h>
+#include <js/String.h>
 #include <js/TypeDecls.h>
 #include <js/Utility.h>  // for UniqueChars
 #include <js/Value.h>
 #include <js/ValueArray.h>
 #include <js/Warnings.h>
-#include <jsapi.h>  // for JS_DefinePropertyById, JS_GetProp...
+#include <jsapi.h>  // for JS_NewPlainObject, JS_NewObject
 
 #include "gi/arg.h"
 #include "gi/boxed.h"
@@ -43,6 +49,7 @@
 #include "cjs/context-private.h"
 #include "cjs/global.h"
 #include "cjs/jsapi-util.h"
+#include "cjs/macros.h"
 #include "cjs/module.h"
 #include "util/log.h"
 
@@ -119,16 +126,16 @@ static bool resolve_namespace_object(JSContext* context,
     JS::RootedObject gi_namespace(context,
                                   gjs_create_ns(context, ns_name.get()));
 
+    JS::RootedValue override(context);
+    if (!lookup_override_function(context, ns_id, &override))
+        return false;
+
     /* Define the property early, to avoid reentrancy issues if
        the override module looks for namespaces that import this */
     if (!JS_DefinePropertyById(context, repo_obj, ns_id, gi_namespace,
                                GJS_MODULE_PROP_FLAGS))
         return false;
 
-    JS::RootedValue override(context);
-    if (!lookup_override_function(context, ns_id, &override))
-        return false;
-
     JS::RootedValue result(context);
     if (!override.isUndefined() &&
         !JS_CallFunctionValue (context, gi_namespace, /* thisp */
@@ -156,7 +163,7 @@ repo_resolve(JSContext       *context,
              JS::HandleId     id,
              bool            *resolved)
 {
-    if (!JSID_IS_STRING(id)) {
+    if (!id.isString()) {
         *resolved = false;
         return true; /* not resolved, but no error */
     }
@@ -368,7 +375,8 @@ gjs_define_info(JSContext       *context,
             } else if (g_type_is_a (gtype, G_TYPE_OBJECT)) {
                 JS::RootedObject ignored1(context), ignored2(context);
                 if (!ObjectPrototype::define_class(context, in_object, info,
-                                                   gtype, &ignored1, &ignored2))
+                                                   gtype, nullptr, 0, &ignored1,
+                                                   &ignored2))
                     return false;
             } else if (G_TYPE_IS_INSTANTIATABLE(gtype)) {
                 JS::RootedObject ignored(context);
@@ -399,7 +407,8 @@ gjs_define_info(JSContext       *context,
             return false;
         break;
     case GI_INFO_TYPE_UNION:
-        if (!gjs_define_union_class(context, in_object, (GIUnionInfo*) info))
+        if (!UnionPrototype::define_class(context, in_object,
+                                          (GIUnionInfo*)info))
             return false;
         break;
     case GI_INFO_TYPE_ENUM:
@@ -477,7 +486,7 @@ gjs_lookup_namespace_object(JSContext  *context,
     }
 
     JS::RootedId ns_name(context, gjs_intern_string_to_id(context, ns));
-    if (ns_name == JSID_VOID)
+    if (ns_name.isVoid())
         return nullptr;
     return gjs_lookup_namespace_object_by_name(context, ns_name);
 }
@@ -547,9 +556,12 @@ lookup_override_function(JSContext             *cx,
         goto fail;
     }
 
+    // If the override module is present, it must have a callable _init(). An
+    // override module without _init() is probably unintentional. (function
+    // being undefined means there was no override module.)
     if (!gjs_object_require_property(cx, module, "override module",
                                      atoms.init(), function) ||
-        !function.isObjectOrNull()) {
+        !function.isObject() || !JS::IsCallable(&function.toObject())) {
         gjs_throw(cx, "Unexpected value for _init in overrides module");
         goto fail;
     }
@@ -720,5 +732,5 @@ JSObject* gjs_new_object_with_generic_prototype(JSContext* cx,
     if (!proto)
         return nullptr;
 
-    return JS_NewObjectWithGivenProto(cx, JS_GetClass(proto), proto);
+    return JS_NewObjectWithGivenProto(cx, JS::GetClass(proto), proto);
 }
