@@ -48,6 +48,16 @@ GJS_EXPORT
 void gjs_list_store_sort(GListStore *store, GjsCompareDataFunc compare_func,
                          void *user_data);
 
+GJS_EXPORT
+GObject* gjs_gtk_custom_sorter_new(GjsCompareDataFunc sort_func,
+                                   void* user_data, GDestroyNotify destroy);
+
+GJS_EXPORT
+void gjs_gtk_custom_sorter_set_sort_func(GObject* sorter,
+                                         GjsCompareDataFunc sort_func,
+                                         void* user_data,
+                                         GDestroyNotify destroy);
+
 /**
  * GjsGLogWriterFunc:
  * @level: the log level
@@ -131,6 +141,30 @@ GBinding* gjs_g_object_bind_property_full(
     GDestroyNotify to_notify, GjsBindingTransformFunc from_callback,
     void* from_data, GDestroyNotify from_notify);
 
+#if GLIB_CHECK_VERSION(2, 72, 0)
+/**
+ * gjs_g_binding_group_bind_full:
+ * @source:
+ * @source_property:
+ * @target:
+ * @target_property:
+ * @flags:
+ * @to_callback: (scope notified) (nullable):
+ * @to_data: (closure to_callback):
+ * @to_notify: (destroy to_data):
+ * @from_callback: (scope notified) (nullable):
+ * @from_data: (closure from_callback):
+ * @from_notify: (destroy from_data):
+ */
+GJS_EXPORT
+void gjs_g_binding_group_bind_full(
+    GBindingGroup* source, const char* source_property, GObject* target,
+    const char* target_property, GBindingFlags flags,
+    GjsBindingTransformFunc to_callback, void* to_data,
+    GDestroyNotify to_notify, GjsBindingTransformFunc from_callback,
+    void* from_data, GDestroyNotify from_notify);
+#endif
+
 /* For imports.overrides.Gtk */
 GJS_EXPORT
 void gjs_gtk_container_child_set_property(GObject* container, GObject* child,
